/* convrenesaslib.c - Convert Renesas library file into GNU archive.
   Copyright 2006
   Free Software Foundation, Inc.

   This file is part of GNU Binutils.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.  */

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include "bfd.h"
#include "bfdver.h"
#include "bucomm.h"
#include "libiberty.h"

/* This is the maximum size of object file data, that the
   buffer can store at a time.  */
#define BUFSIZE 		8192
/* Macro for Renesas library file: Little endian format.  */
#define RLIB_LITTLE_ENDIAN	0x01
/* Macro for Renesas library file: Big endian format.  */
#define RLIB_BIG_ENDIAN		0x02
/* Macro for Renesas library file: SH target.  */
#define RLIB_SH_CPU		0x01
/* ar utility */
#define AR                      "ar"
/* objcopy utility */
#define OBJCOPY                 "objcopy"
/* convrenesaslib utility */
#define CONVRENESASLIB          "convrenesaslib"

void process_renesas_library_file (const char* filename);
char *program_name;
char *invoke_utility_name;
int remove_underscore = 0;
int verbose = 0;

/* Used to hold Renesas library header information.  */
struct rlib_header
{
  unsigned char magic_num [4];
  union
    {
      unsigned char lib_cpu_endian_info [4];
      unsigned int info;
    } cpu;
  union
    {
      unsigned char num_mod [4];
      unsigned int num_modules;
    } num_of_modules;
  union
    {
      unsigned char string_tbl_offset [4];
      unsigned int string_table_offset;
    } string_table_file_offset;
  union
    {
      unsigned char table_size [4];
      unsigned int size;
    } string_table_size;
};


/* Used to hold module header information.  */
struct module_header
{
  union
    {
      unsigned char name_offset [4];
      unsigned int symb_name_offset;
    } symbol_name_offset;
  union
    {
      unsigned char name_string_size [4];
      unsigned int symb_name_string_size;
    } symbol_name_string_size;
  union
    {
      unsigned char module_offset [4];
      unsigned int mod_file_offset;
    } module_file_offset;
  union
    {
      unsigned char mod_size [4];
      unsigned int size;
    } module_size;
  unsigned int ignore_data;
};

static void
usage (FILE *stream, int status)
{
  fprintf (stream, _("Usage: %s [option] [file(s)] \n"), program_name);
  fprintf (stream, _(" Convert Renesas SH library file to GNU library archive with '.rlib' extension.\n"));
  fprintf (stream, _(" This utility supports only Renesas SH targets\n"));
  fprintf (stream, _(" The options are:\n\
  --remove-underscore: Remove the leading underscore for all global symbols\n\
  --help:              Display this information\n\
  --version:           Display this program's version number\n\
  --verbose:           Verbosely list files processed\n\
  Supported targets:   elf32-sh elf32-shl\n\
  \n"));

  if (status == 0)
    fprintf (stream, _("Report bugs to %s\n"), REPORT_BUGS_TO);
  exit (status);
}


/* Option specification for getopt_long.  */
static const struct option long_options[] =
{
  { "help",	no_argument, NULL, 'h' },
  { "version", no_argument, NULL, 'p' },
  { "verbose", no_argument, NULL, 'v' },
  { "remove-underscore", no_argument, NULL, 'r' },
  { NULL, no_argument, NULL, 0   },
};


/* The main routine.  */
int
main(int argc, char **argv)
{
  int count;
  int opt_char;
  int initial_value = 1;

  program_name = (char*) xmalloc (strlen (argv [0]) * sizeof (char));
  invoke_utility_name = (char*) xmalloc ((strlen (argv [0])) * sizeof (char));
  program_name = argv [0];

  /* Parse options.  */
  do
    {
      opt_char = getopt_long (argc, argv, "hpvr", long_options, NULL);
      switch (opt_char)
        {
        case '?':  /* Unrecognized option.  */
          usage (stderr, 1);
          break;

        case 'h':
          usage (stdout, 0);
          break;

        case 'p':
          /* This output is intended to follow the GNU standards document.  */
          printf (_("GNU convrenesaslib %s\n"), BFD_VERSION_STRING);
          printf (_("Copyright 2006 Free Software Foundation, Inc.\n"));
          printf (_("\
This program is free software; you may redistribute it under the terms of\n\
the GNU General Public License.  This program has absolutely no warranty.\n"));
          exit (EXIT_SUCCESS);;
          break;

        case 'v':
          verbose = 1;
          initial_value = 2;
          break;

        case 'r':
          remove_underscore = 1;
          initial_value = 2;
          break;
        }
    }
  while (opt_char != -1);

  if (argc < 2)
    usage (stdout, 0);

  switch (argc)
    {
    case 2:
      if ((remove_underscore == 1) || (verbose ==1))
        usage (stderr, 1);
      break;

    case 3:
      if ((remove_underscore == 1) && (verbose ==1))
        usage (stderr, 1);
      break;

    default:
      break;
    }

  if ((remove_underscore == 1) && (verbose == 1))
    {
      remove_underscore = 1;
      initial_value = 3;
      verbose = 1;
    }


  for (count = initial_value; count < argc; count++)
    process_renesas_library_file (argv [count]);
  return 0;
}

void
process_renesas_library_file (const char* filename)
{
  FILE *fbopen, *fbwrite;
  char c;
  /* Array of 5 bytes for magic number "HLIB" + '\0'.  */
  unsigned char magic_num_arr [5];
  char *module_name;
  char *filepath;
  char *objbuf = xmalloc (BUFSIZE * sizeof (char));
  int length, nread;
  int size, ncopied, tocopy;
  int hdr_offset = 0;
  int arg_count = 0;
  char** files;
  unsigned int count;
  int i=0;
  char *ptr;

  struct rlib_header rlib_hdr, rlib_hdr_big;
  struct module_header module_hdr, module_hdr_big;
  char *arguments;
  char *cmdline;
  int cmdline_length = 0;
  size = ncopied = tocopy = 0;
  nread = 0;
  count = 0;
  length = 0;

  /* Open the input Renesas library file.  */
  fbopen = fopen (filename, "rb");
  if (fbopen == NULL)
    {
      perror (filename);
      xexit (1);
    }

  /*
   * Renesas library header information
   * magic number :		rlib_hdr.magic_num
   * cpu lib info :		rlib_hdr.cpu.lib_cpu_endian_info[0]
   * cpu cpu info :		rlib_hdr.cpu.lib_cpu_endian_info[1]
   * cpu endian info :		rlib_hdr.cpu.lib_cpu_endian_info[2]
   * num_of_modules :		rlib_hdr.num_of_modules.num_modules
   * string_table_file_offset :	rlib_hdr.string_table_file_offset.string_table_offset
   * string_table_size :	rlib_hdr.string_table_size.size
  */

  /* Read the Renesas library header.  */
  nread = fread (&rlib_hdr, sizeof(rlib_hdr), 1, fbopen);

  /* Terminate magic number with '\0' for error checking.  */
  strcpy (magic_num_arr, rlib_hdr.magic_num);
  magic_num_arr[4] = '\0';

  if ((strcmp (magic_num_arr, "HLIB") != 0) || (rlib_hdr.cpu.lib_cpu_endian_info[1] != RLIB_SH_CPU))
    {
      /* Close file pointer.  */
      fclose (fbopen);
      free (objbuf);
      /* Not Renesas magic bytes, not Renesas archive.  */
      fatal (_("\n%s is not a Renesas SH target library file\n convrenesaslib utility supports only Renesas SH targets\n"), filename);
    }

  /* Big endian.  */
  if (rlib_hdr.cpu.lib_cpu_endian_info [2] == RLIB_BIG_ENDIAN)
    {
      /* Point to start location of file.  */
      fseek (fbopen, 0, SEEK_SET);
      nread = fread (&rlib_hdr_big, sizeof (rlib_hdr_big), 1, fbopen);

      /* Populate the parameters of rlib_hdr with values in rlib_hdr_big: endian related.  */

      /* Renesas library file information: Number of modules.  */
      rlib_hdr.num_of_modules.num_mod [0] = rlib_hdr_big.num_of_modules.num_mod [3];
      rlib_hdr.num_of_modules.num_mod [1] = rlib_hdr_big.num_of_modules.num_mod [2];
      rlib_hdr.num_of_modules.num_mod [2] = rlib_hdr_big.num_of_modules.num_mod [1];
      rlib_hdr.num_of_modules.num_mod [3] = rlib_hdr_big.num_of_modules.num_mod [0];

      /* Renesas library file information: String table file offset.  */
      rlib_hdr.string_table_file_offset.string_tbl_offset [0] = rlib_hdr_big.string_table_file_offset.string_tbl_offset [3];
      rlib_hdr.string_table_file_offset.string_tbl_offset [1] = rlib_hdr_big.string_table_file_offset.string_tbl_offset [2];
      rlib_hdr.string_table_file_offset.string_tbl_offset [2] = rlib_hdr_big.string_table_file_offset.string_tbl_offset [1];
      rlib_hdr.string_table_file_offset.string_tbl_offset [3] = rlib_hdr_big.string_table_file_offset.string_tbl_offset [0];
    }

  /* Extra 3 bytes for files [0] files [1] files [2].  */
  files = (char**) xmalloc ((rlib_hdr.num_of_modules.num_modules + 4) * sizeof (char*));

  files [0] = (char*) xmalloc(1 * sizeof (char));
  /* Holds the options "-qc" for ar. 1 byte for NULL terminator.  */
  files [1] = (char*) xmalloc((strlen ("-qc") + 1) * sizeof (char));
  /* '.rlib' will be appended to the filename. 1 byte for NULL terminator.  */
  files [2] = (char*) xmalloc((strlen (filename) + strlen (".rlib") + 1) * sizeof (char));

  /* Object files are listed from count number 3 in the array files[].  */
  arg_count = 3;
  filepath = (char*) xmalloc ((strlen(filename) + strlen (".rlib") + 1) * sizeof (char));
  strcpy (filepath, filename);
  strcat (filepath,".rlib");

  strcpy (files [0], "");
  strcpy (files [1], "-qc");
  strcpy (files [2], filepath);

  free (filepath);

  /* 'files[1] files[2]': length of this command line.  */
  cmdline_length = strlen (files [1]) + strlen (files [2]);

  if (verbose == 1)
    printf(_("\n Extracted object file(s) from %s "), filename);

  strcpy (invoke_utility_name, program_name);

  /* Replace the name 'convrenesaslib' (the name has 14 bytes) in invoke_utility_name with ar.  */
  ptr = strrchr (invoke_utility_name, 'c');
  strcpy (ptr, "");
  strcat (invoke_utility_name, AR);

  /* Extract object files from the Renesas library file.  */
  for (count = 0; count < rlib_hdr.num_of_modules.num_modules; count++)
    {
      /* Point to respective header for a module.  */
      hdr_offset = sizeof (rlib_hdr) + (count * sizeof (module_hdr));
      /* Point to start location of object header file.  */
      fseek (fbopen, hdr_offset, SEEK_SET);

      /*
       * Object module header information
       * symbol_name_offset :	  module_hdr.symbol_name_offset.symb_name_offset
       * symbol_name_string_size : module_hdr.symbol_name_string_size.symb_name_string_size
       * module_file_offset :	  module_hdr.module_file_offset.mod_file_offset
       * module_size :		  module_hdr.module_size.size
       * ignore_data :		  module_hdr.ignore_data
      */

      /* Read Module header.  */
      nread = fread (&module_hdr, sizeof (module_hdr), 1, fbopen);

      /* For Renesas library file: big endian format.  */
      if (rlib_hdr.cpu.lib_cpu_endian_info [2] == RLIB_BIG_ENDIAN)
        {
          fseek (fbopen, hdr_offset, SEEK_SET);
	  nread = fread (&module_hdr_big, sizeof (module_hdr_big), 1, fbopen);

	  /* Renesas library file: Symbol name offset of module within string table */
	  module_hdr.symbol_name_offset.name_offset [0] = module_hdr_big.symbol_name_offset.name_offset [3];
	  module_hdr.symbol_name_offset.name_offset [1] = module_hdr_big.symbol_name_offset.name_offset [2];
	  module_hdr.symbol_name_offset.name_offset [2] = module_hdr_big.symbol_name_offset.name_offset [1];
	  module_hdr.symbol_name_offset.name_offset [3] = module_hdr_big.symbol_name_offset.name_offset [0];

	  /* Renesas Library file: Module information's file offset.  */
	  module_hdr.module_file_offset.module_offset [0] = module_hdr_big.module_file_offset.module_offset [3];
	  module_hdr.module_file_offset.module_offset [1] = module_hdr_big.module_file_offset.module_offset [2];
	  module_hdr.module_file_offset.module_offset [2] = module_hdr_big.module_file_offset.module_offset [1];
	  module_hdr.module_file_offset.module_offset [3] = module_hdr_big.module_file_offset.module_offset [0];

	  /* Renesas library file: Module size information.  */
	  module_hdr.module_size.mod_size [0] = module_hdr_big.module_size.mod_size [3];
	  module_hdr.module_size.mod_size [1] = module_hdr_big.module_size.mod_size [2];
	  module_hdr.module_size.mod_size [2] = module_hdr_big.module_size.mod_size [1];
	  module_hdr.module_size.mod_size [3] = module_hdr_big.module_size.mod_size [0];
        }

      /* Size of object module.  */
      size = module_hdr.module_size.size;
      if (size < 0)
        {
          /* Close file pointer.  */
          fclose (fbopen);
          /* Free allocated memory.  */
          free (files);
          fatal (_("\n stat returns negative size for %s \n"), filename);
        }

      /* Renesas library file: Name of object file to be extracted */
      fseek (fbopen, (rlib_hdr.string_table_file_offset.string_table_offset + module_hdr.symbol_name_offset.symb_name_offset), SEEK_SET);
      length = 0;
      c = fgetc (fbopen);
      while (c != '\0')
        {
          /* Length of module name.  */
	  length++;
	  c = fgetc (fbopen);
        }
      length++;
      /* Allocate memory for module_name.  */
      module_name = (char*) xmalloc ((length + strlen (".robj") + 1) * sizeof (char));
      /* Allocate memory for filepath.  */
      filepath = (char*) xmalloc ((length + strlen (".robj") + 1) * sizeof (char));
      /* Initialize the buffers.  */
      strcpy (filepath, "");
      strcpy (module_name, "");

      /* Renesas library file: Offset for object module name in string table.  */
      fseek (fbopen, (rlib_hdr.string_table_file_offset.string_table_offset + module_hdr.symbol_name_offset.symb_name_offset), SEEK_SET);
      /* Name of the module is stored in 'module_name'.  */
      nread = fread (module_name, length, 1, fbopen);

      /* Name of object file to be created.  */
      strcat (module_name, ".robj");
      strcpy (filepath, module_name);

      /* Length of the command line string.  */
      cmdline_length = cmdline_length + strlen (filepath);

      if (verbose == 1)
        printf(_("\n %s "), filepath);

      free (module_name);
      /* arguments to be passed to 'create_gnu_archive' function: module names.  */
      files[arg_count] = (char*) xmalloc ((strlen (filepath) + 1) * sizeof (char));

      strcpy (files[arg_count], filepath);
      arg_count++;

      /* Extract object file.  */

      /* Start location of object file.  */
      fseek (fbopen, module_hdr.module_file_offset.mod_file_offset, SEEK_SET);
      ncopied = 0;

      if (size > 0)
        {
          fbwrite = fopen (filepath, "wb");
	  if (fbwrite == NULL)
	    {
	      perror (filepath);
              free (filepath);
              xexit (1);
	    }
	  
      	  /* read 52 bytes of elf header into buffer  */
	  nread = fread (objbuf, 1, 52, fbopen);
	   /* Modify 'e_flag'(processor specific flag) value in each Renesas object file to that expected by GNUSH 
	    * for CPUs sh2a and sh2a-fpu  */
	  if(objbuf[39] == 0x04) objbuf[39] = 0x13;
	  if(objbuf[39] == 0x05) objbuf[39] = 0x0d;
	  fwrite (objbuf, 1, 52, fbwrite);
	  ncopied = 52;

	  while (ncopied < size)
	    {
	      tocopy = size - ncopied;
	      if (tocopy > BUFSIZE)
	        tocopy = BUFSIZE;

	      /* Copy object module from library file.  */
	      nread = fread (objbuf, 1, tocopy, fbopen);
	      if (nread == tocopy)
	        {
	          fwrite (objbuf, 1, tocopy, fbwrite);
	          ncopied = ncopied + tocopy;
	        }
	      else
	        {
	          /* Close file pointer.  */
	          fclose (fbopen);
	          fclose (fbwrite);
	          /* Delete extracted object files.  */
                  /* Object files are listed from count number 3 in the array files[].  */
	          for (i = 3; i < (arg_count - 1); i++)
	            remove (files[i]);
	          /* Free allocated memory.  */
	          free (files);
                  free (filepath);
	          fatal (_("\n %s is not a valid archive \n"), filename);
	        }
	    }
          fclose (fbwrite);
        }
      else if (size == 0)
        {
          remove (filepath);
	  fbwrite = fopen (filepath, "wb");
	  if (fbwrite == NULL)
	    {
	      perror (filepath);
              free (filepath);
              xexit (1);
	    }
          fclose (fbwrite);
        }
    }
  fclose (fbopen);
  free (objbuf);
  free (filepath);

  /* Delete if any file with the existing name.  */
  remove (files[2]);

  /* Allocate space for arguments to be passed to ar. Length
  of the parameter names + spaces in the command line. 4 extra to prevent malloced
  area from getting corrupted. Somehow it is getting corrupted.  */
  arguments = (char *) xmalloc ((cmdline_length + rlib_hdr.num_of_modules.num_modules + 4) * sizeof (char));
  strcpy (arguments, " ");
  for (i = 1; i < arg_count; i++)
    {
      strcat (arguments, files [i]);
      strcat (arguments, " ");
    }

  /* Allocate space for the command line to be formed for invoking ar. Length
  of the parameter names + spaces in the command line. 4 extra to prevent malloced
  area from getting corrupted. Somehow it is getting corrupted.  */
  cmdline = (char *) xmalloc ((strlen (arguments) + strlen (invoke_utility_name) + 4) * sizeof (char));
  strcpy (cmdline, invoke_utility_name);
  strcat (cmdline, " ");
  strcat (cmdline, arguments);

  /* Invoke the 'ar' utility.  */
  if (verbose == 1)
    {
      printf(_("\n Creating GNU archive %s "), files [2]);
      printf(_("\n %s \n"), cmdline);
    }

  /* Invoking the ar utility with command line parameters.  */
  system(cmdline);

  /* Invoke the objcopy utility if the option '--remove-underscore' has been passed.  */
  if (remove_underscore == 1)
    {
      i = strlen (invoke_utility_name);

      /* Replace the name 'ar' (2 bytes) in invoke_utility_name with objcopy */
      invoke_utility_name [strlen (invoke_utility_name) - strlen (AR)] = '\0';
      strcat (invoke_utility_name, OBJCOPY);

      strcpy (cmdline, invoke_utility_name);
      strcat (cmdline, " ");
      strcat (cmdline, "--remove-leading-char");
      strcat (cmdline, " ");
      strcat (cmdline, files [2]);

      /* Invoke the 'objcopy' utility.  */
      if (verbose == 1)
        {
          printf(_("\n Remove leading underscores in %s "), files [2]);
          printf(_("\n %s \n"), cmdline);
        }

      /* Invoking objcopy utility with command line parameters.  */
      system (cmdline);
    }

  if (verbose == 1)
    printf(_("\n Deleting object file(s) extracted from %s "), filename);

  /* Object files are listed from count number 3 in the array files[].  */
  for (i = 3; i < arg_count; i++)
    {
      remove (files[i]);
      if (verbose == 1)
        printf(_("\n %s"), files[i]);
    }

  if (verbose == 1)
    printf(_("\n "));
  free (files);
  free (arguments);
  free (cmdline);
}



